"""
CV-ArtSmartObject v2.95
Copyright: 2012-2013 Rick Barrett
Written for CINEMA 4D R13+
"""

import os
import sys
import subprocess
import pickle
import webbrowser

import c4d

__currdir__ = os.path.join(os.path.dirname(__file__),'lib')
if __currdir__ not in sys.path:
    sys.path.insert(0, __currdir__)

import vecfile, pdf, svg, geom
import artsmartcore

#=====================================================================================================================#
#   SYMBOLS
#=====================================================================================================================#
try:
    import symbol_parser
    c4d_symbol_file = os.path.join(os.path.dirname(__file__), "res")
    symbol_parser.parse_and_export_in_caller(c4d_symbol_file)
except ImportError:
    ART_FILE		=30000
    ART_LAYEROFFSET		=30001
    ART_DEPTH		=30002
    ART_CONTENTS		=30003
    ART_RECENTER		=30004
    ART_PATHOFFSET		=30005
    ART_USEMOGRAPH		=30006
    ART_CINEWARE		=30007
    ART_FILTERGROUP		=30010
    ART_LAYERFILTER		=30011
    ART_PATHFILTER_ON	=30012
    ART_PATHFILTER		=30013
    ART_RECENTER_Z		=30014
    ART_FORCEPDF		=30015
    ART_STROKESETTINGS	=30016
    ART_STROKESCALE		=30017
    ART_STROKEASPECT	=30018
    ART_STROKEROUNDING	=30019
    ART_STROKEGROWTH	=30020
    ART_STROKEALIGN		=30021
    ART_STROKEALIGN_START = 30022
    ART_STROKEALIGN_END = 30023
    ART_STROKEALIGN_CENTER = 30024
    ART_STROKEALIGN_STARTEND = 30025
    ART_STROKEOFFSET = 30026
    ART_GREEDYFILL = 30050
    ART_AUTOCLOSE = 30051
    ART_LEGACYCENTER = 30052
    ART_PATHEFFECTOR	=30100
    ART_PATHEFFECTOR_LIST = 30101
    ART_LAYEREFFECTOR   =30200
    ART_LAYEREFFECTOR_LIST = 30201
    ART_RELOAD		=5000
    ART_OPEN		=5001
    ART_CREATEMATERIALS		=5002
    ART_INFO		=5005
    ART_DEBUGMSG	=5006
    ART_CREATEOBJECT = 6000
    ART_CHECKSUM		=9000
    ART_STROKED			=9001
    ART_FILLED			=9002
    ART_STROKECOLOR		=9003
    ART_FILLCOLOR		=9004
    ART_STROKEWIDTH		=9005

VERSION = 2.98

PLUGIN_ID = 1028655
ICON_ID = 1028659
FRACTURE_ID = 1018791
TAG_ID = 1028660
DEBUG = False

class CVArtSmart(c4d.plugins.ObjectData):
    
    def __init__(self):
        self.SetOptimizeCache(True)
        self.art = None
        self.z = 0
        self.fn = ""
        self.matGroup = None
        self.userdataIdx = {}
        self.matIdx = {}
        self.rebuildUserdata = False
        self.pathControl = True
        self.forcePdf = False
        self.greedyFill = False
        self.loaded = False
        
    def Init(self, node):
        #load attributes from .h into c4d. .res describes the structure
        self.InitAttr(host=node, type=str, id=[ART_FILE])
        self.InitAttr(host=node, type=float, id=[ART_LAYEROFFSET])
        self.InitAttr(host=node, type=float, id=[ART_PATHOFFSET])
        self.InitAttr(host=node, type=float, id=[ART_DEPTH])
        self.InitAttr(host=node, type=bool, id=[ART_RECENTER])
        self.InitAttr(host=node, type=bool, id=[ART_RECENTER_Z])
        self.InitAttr(host=node, type=bool, id=[ART_USEMOGRAPH])
        self.InitAttr(host=node, type=bool, id=[ART_CINEWARE]) 
        self.InitAttr(host=node, type=bool, id=[ART_FORCEPDF])
        self.InitAttr(host=node, type=bool, id=[ART_GREEDYFILL])
        self.InitAttr(host=node, type=bool, id=[ART_AUTOCLOSE])
        self.InitAttr(host=node, type=bool, id=[ART_LEGACYCENTER])
        self.InitAttr(host=node, type=float, id=[c4d.CAP_STARTRADIUS])
        self.InitAttr(host=node, type=float, id=[c4d.CAP_ENDRADIUS])
        self.InitAttr(host=node, type=float, id=[c4d.CAP_REGULARWIDTH])
        self.InitAttr(host=node, type=float, id=[ART_STROKESCALE])
        self.InitAttr(host=node, type=float, id=[ART_STROKEASPECT])
        self.InitAttr(host=node, type=float, id=[ART_STROKEROUNDING])
        self.InitAttr(host=node, type=float, id=[ART_STROKEGROWTH])
        #self.InitAttr(host=node, type=int, id=[ART_STROKEALIGN])
        self.InitAttr(host=node, type=float, id=[ART_STROKEOFFSET])
        #self.InitAttr(host=node, type=int, id=[c4d.SPLINEOBJECT_INTERPOLATION])
        #self.InitAttr(host=node, type=int, id=[c4d.SPLINEOBJECT_SUB])
        #self.InitAttr(host=node, type=float, id=[c4d.SPLINEOBJECT_ANGLE])
        #self.InitAttr(host=node, type=float, id=[c4d.SPLINEOBJECT_MAXIMUMLENGTH])
        self.InitAttr(host=node, type=str, id=[ART_DEBUGMSG])
        
        node[ART_FILE] = ""
        node[ART_LAYEROFFSET] = 0
        node[ART_PATHOFFSET] = 10        
        node[ART_DEPTH] = 10
        node[ART_RECENTER] = True
        node[ART_RECENTER_Z] = False
        node[ART_USEMOGRAPH] = False
        node[ART_FORCEPDF] = False
        node[ART_GREEDYFILL] = False
        node[ART_AUTOCLOSE] = False
        node[ART_LEGACYCENTER] = True
        node[ART_CINEWARE] = False

        node[c4d.CAPSANDBEVELS_CAP_ENABLE_START] = True
        node[c4d.CAPSANDBEVELS_CAP_ENABLE_END] = True
        node[c4d.CAPSANDBEVELS_AVOIDINTERSECTION] = True
        node[c4d.CAPSANDBEVELS_CAP_TYPE] = 2
        node[c4d.CAPSANDBEVELS_PHONGBREAKROUNDING] = True
        node[c4d.CAPSANDBEVELS_STARTBEVEL_TYPE] = c4d.CAPSANDBEVELS_BEVEL_TYPE_ROUND
        node[c4d.CAPSANDBEVELS_STARTBEVEL_OFFSET] = 0
        node[c4d.CAPSANDBEVELS_STARTBEVEL_DEPTH] = 1.0
        node[c4d.CAPSANDBEVELS_ENDBEVEL_TYPE] = c4d.CAPSANDBEVELS_BEVEL_TYPE_ROUND
        node[c4d.CAPSANDBEVELS_ENDBEVEL_OFFSET] = 0
        node[c4d.CAPSANDBEVELS_ENDBEVEL_DEPTH] = 1.0
        node[c4d.CAPSANDBEVELS_STARTBEVEL_SEGMENTS] = 3
        node[c4d.CAPSANDBEVELS_ENDBEVEL_SEGMENTS] = 3

        node[c4d.CAP_START] = c4d.CAPTYPE_CAP
        node[c4d.CAP_STARTRADIUS] = 5
        node[c4d.CAP_END] = c4d.CAPTYPE_CAP
        node[c4d.CAP_ENDRADIUS] = 5
        node[c4d.CAP_ROUNDING] = c4d.CAP_ROUNDING_CONVEX
        node[c4d.CAP_HULLINWARD] = True
        node[c4d.CAP_TYPE] = c4d.CAP_TYPE_NGON
        node[c4d.CAP_REGULARWIDTH] = 10
        
        node[ART_STROKESCALE] = 1
        node[ART_STROKEASPECT] = 10
        node[ART_STROKEROUNDING] = 0
        node[ART_STROKEGROWTH] = 1
        node[ART_STROKEOFFSET] = 0
                
        node[c4d.SPLINEOBJECT_INTERPOLATION] = c4d.SPLINEOBJECT_INTERPOLATION_ADAPTIVE
        node[c4d.SPLINEOBJECT_SUB] = 8
        node[c4d.SPLINEOBJECT_ANGLE] = c4d.utils.Rad(5)
        node[c4d.SPLINEOBJECT_MAXIMUMLENGTH] = 5
        
        node[ART_DEBUGMSG] = ""
        
        self.IndexUserData(node)
        
        return True
    
    def Read(self,node,hf,level):
        artDill = hf.ReadString()
        self.art = pickle.loads(artDill)
        self.fn = node[ART_FILE]
        return True
        
    def Write(self,node,hf):
        artDill = pickle.dumps(self.art)
        hf.WriteString(artDill)
        return True

    #CopyTo is broken in 17.000 - 17.044
    if c4d.GetC4DVersion() < 17000 or c4d.GetC4DVersion() > 17045:
        def CopyTo(self,dest,snode,dnode,flags,trn):
             dest.art = self.art
             dest.fn = self.fn
             dest.matGroup = self.matGroup
             return True
    
    if c4d.GetC4DVersion() >= 15043:
        # This function can't be included with earlier versions of C4D due to bug #107686 / BitmapButton
        def GetDEnabling(self,node,id,t_data,flags,itemdesc):
            if id[0].id == ART_USEMOGRAPH:
                return c4d.modules.CheckMoGraph()
            elif id[0].id == c4d.SPLINEOBJECT_SUB:
                if node[c4d.SPLINEOBJECT_INTERPOLATION] == c4d.SPLINEOBJECT_INTERPOLATION_NATURAL or \
                   node[c4d.SPLINEOBJECT_INTERPOLATION] == c4d.SPLINEOBJECT_INTERPOLATION_UNIFORM:
                   return True
                else:
                   return False
            elif id[0].id == c4d.SPLINEOBJECT_ANGLE:
                if node[c4d.SPLINEOBJECT_INTERPOLATION] == c4d.SPLINEOBJECT_INTERPOLATION_ADAPTIVE or \
                   node[c4d.SPLINEOBJECT_INTERPOLATION] == c4d.SPLINEOBJECT_INTERPOLATION_SUBDIV:
                   return True
                else:
                   return False
            elif id[0].id == c4d.SPLINEOBJECT_MAXIMUMLENGTH:
                return node[c4d.SPLINEOBJECT_INTERPOLATION] == c4d.SPLINEOBJECT_INTERPOLATION_SUBDIV
            elif id[0].id >= c4d.CAP_START and id[0].id <= c4d.CAP_TYPE:
                #return node[ART_DEPTH] > 0
                return True
            elif id[0].id == ART_LAYEREFFECTOR_LIST or id[0].id == ART_PATHEFFECTOR_LIST:
                return bool(node[ART_USEMOGRAPH])

            elif id[0].id == c4d.CAPSANDBEVELS_STARTBEVEL_HEIGHT:
                return bool(node[c4d.CAPSANDBEVELS_STARTBEVEL_MODE])
            elif id[0].id == c4d.CAPSANDBEVELS_ENDBEVEL_HEIGHT:
                return bool(node[c4d.CAPSANDBEVELS_ENDBEVEL_MODE])
            elif id[0].id == c4d.CAPSANDBEVELS_STARTBEVEL_DEPTH:
                return bool(node[c4d.CAPSANDBEVELS_STARTBEVEL_OFFSET]>0)
            elif id[0].id == c4d.CAPSANDBEVELS_ENDBEVEL_DEPTH:
                return bool(node[c4d.CAPSANDBEVELS_ENDBEVEL_OFFSET]>0)
            elif id[0].id == c4d.CAPSANDBEVELS_STARTBEVEL_ABSOLUTEDEPTH:
                return bool(node[c4d.CAPSANDBEVELS_STARTBEVEL_ABSOLUTE])
            elif id[0].id == c4d.CAPSANDBEVELS_ENDBEVEL_ABSOLUTEDEPTH:
                return bool(node[c4d.CAPSANDBEVELS_ENDBEVEL_ABSOLUTE])

            return True

    def descHelper(self, node, description, param, criteria, changeID=c4d.DESC_HIDE):
        # Get description single ID
        singleID = description.GetSingleDescID()

        # Check if dynamic parameter with ID paramID has to be added
        paramDID = c4d.DescID(c4d.DescLevel(param))

        if singleID is None or paramDID.IsPartOf(singleID)[0]:
            bc = description.GetParameterI(paramDID, None)
            bc[changeID] = criteria

    if c4d.GetC4DVersion() > 20000:
        def GetDDescription(self, node, description, flags):
            # Before adding dynamic parameters, load the parameters from the description resource
            if not description.LoadDescription(node.GetType()):
                return False

            # Get description single ID
            singleID = description.GetSingleDescID()

            # Check if dynamic parameter with ID paramID has to be added
            startGroup = c4d.DescID(c4d.DescLevel(c4d.CAPSANDBEVELS_STARTBEVEL_GROUP))

            if singleID is None or startGroup.IsPartOf(singleID)[0]:
                bc = description.GetParameterI(startGroup, None)
                if node[c4d.CAPSANDBEVELS_BEVEL_SEPARATE]:
                    bc[c4d.DESC_NAME] = 'Start Bevel'
                else:
                    bc[c4d.DESC_NAME] = 'Both Bevels'

            self.descHelper(node, description, c4d.CAPSANDBEVELS_ENDBEVEL_GROUP, not node[c4d.CAPSANDBEVELS_BEVEL_SEPARATE])
            self.descHelper(node, description, c4d.CAPSANDBEVELS_STARTBEVEL_CURVE, node[c4d.CAPSANDBEVELS_STARTBEVEL_TYPE]!=c4d.CAPSANDBEVELS_BEVEL_TYPE_CURVE)
            self.descHelper(node, description, c4d.CAPSANDBEVELS_ENDBEVEL_CURVE, node[c4d.CAPSANDBEVELS_ENDBEVEL_TYPE]!=c4d.CAPSANDBEVELS_BEVEL_TYPE_CURVE)
            self.descHelper(node, description, c4d.CAPSANDBEVELS_STARTBEVEL_DEPTH,
                                node[c4d.CAPSANDBEVELS_STARTBEVEL_TYPE]!=c4d.CAPSANDBEVELS_BEVEL_TYPE_CURVE and \
                                node[c4d.CAPSANDBEVELS_STARTBEVEL_TYPE]!=c4d.CAPSANDBEVELS_BEVEL_TYPE_ROUND)
            self.descHelper(node, description, c4d.CAPSANDBEVELS_ENDBEVEL_DEPTH,
                                node[c4d.CAPSANDBEVELS_ENDBEVEL_TYPE] != c4d.CAPSANDBEVELS_BEVEL_TYPE_CURVE and \
                                node[c4d.CAPSANDBEVELS_ENDBEVEL_TYPE] != c4d.CAPSANDBEVELS_BEVEL_TYPE_ROUND)
            self.descHelper(node, description, c4d.CAPSANDBEVELS_STARTBEVEL_ALONGNORMALS,
                                node[c4d.CAPSANDBEVELS_STARTBEVEL_TYPE] != c4d.CAPSANDBEVELS_BEVEL_TYPE_CURVE)
            self.descHelper(node, description, c4d.CAPSANDBEVELS_STARTBEVEL_ABSOLUTE,
                                node[c4d.CAPSANDBEVELS_STARTBEVEL_TYPE] != c4d.CAPSANDBEVELS_BEVEL_TYPE_CURVE)
            self.descHelper(node, description, c4d.CAPSANDBEVELS_STARTBEVEL_ABSOLUTEDEPTH,
                                node[c4d.CAPSANDBEVELS_STARTBEVEL_TYPE] != c4d.CAPSANDBEVELS_BEVEL_TYPE_CURVE)
            self.descHelper(node, description, c4d.CAPSANDBEVELS_ENDBEVEL_ALONGNORMALS,
                                node[c4d.CAPSANDBEVELS_ENDBEVEL_TYPE] != c4d.CAPSANDBEVELS_BEVEL_TYPE_CURVE)
            self.descHelper(node, description, c4d.CAPSANDBEVELS_ENDBEVEL_ABSOLUTE,
                                node[c4d.CAPSANDBEVELS_ENDBEVEL_TYPE] != c4d.CAPSANDBEVELS_BEVEL_TYPE_CURVE)
            self.descHelper(node, description, c4d.CAPSANDBEVELS_ENDBEVEL_ABSOLUTEDEPTH,
                                node[c4d.CAPSANDBEVELS_ENDBEVEL_TYPE] != c4d.CAPSANDBEVELS_BEVEL_TYPE_CURVE)
            self.descHelper(node, description, c4d.CAPSANDBEVELS_CAP_DELAUNAY_GRAD,
                                node[c4d.CAPSANDBEVELS_CAP_TYPE] != c4d.CAPSANDBEVELS_CAP_TYPE_DELAUNAY)
            self.descHelper(node, description, c4d.CAPSANDBEVELS_CAP_REGULAR_SIZE,
                                node[c4d.CAPSANDBEVELS_CAP_TYPE] != c4d.CAPSANDBEVELS_CAP_TYPE_REGULAR)
            self.descHelper(node, description, c4d.CAPSANDBEVELS_CAP_DELAUNAY_QUAD,
                                node[c4d.CAPSANDBEVELS_CAP_TYPE] != c4d.CAPSANDBEVELS_CAP_TYPE_DELAUNAY and \
                                node[c4d.CAPSANDBEVELS_CAP_TYPE] != c4d.CAPSANDBEVELS_CAP_TYPE_REGULAR)
            self.descHelper(node, description, c4d.CAPSANDBEVELS_STARTBEVEL_LOADPRESET,
                                True)
            self.descHelper(node, description, c4d.CAPSANDBEVELS_STARTBEVEL_SAVEPRESET,
                                True)
            self.descHelper(node, description, c4d.CAPSANDBEVELS_ENDBEVEL_LOADPRESET,
                                True)
            self.descHelper(node, description, c4d.CAPSANDBEVELS_ENDBEVEL_SAVEPRESET,
                                True)

            # Handle Parent Collapse of Depth attribute
            if node[c4d.CAPSANDBEVELS_STARTBEVEL_TYPE]==c4d.CAPSANDBEVELS_BEVEL_TYPE_CURVE:
                self.descHelper(node, description, c4d.CAPSANDBEVELS_STARTBEVEL_DEPTH,
                                -1, c4d.DESC_PARENT_COLLAPSE)    

            else:
                self.descHelper(node, description, c4d.CAPSANDBEVELS_STARTBEVEL_DEPTH,
                                0, c4d.DESC_PARENT_COLLAPSE)    
 
            if node[c4d.CAPSANDBEVELS_ENDBEVEL_TYPE]==c4d.CAPSANDBEVELS_BEVEL_TYPE_CURVE:
                self.descHelper(node, description, c4d.CAPSANDBEVELS_ENDBEVEL_DEPTH,
                                -1, c4d.DESC_PARENT_COLLAPSE)   
            else:
                self.descHelper(node, description, c4d.CAPSANDBEVELS_ENDBEVEL_DEPTH,
                                0, c4d.DESC_PARENT_COLLAPSE)   

            # self.descHelper(node, description, c4d.CAPSANDBEVELS_CAP_TYPE,
            #                 node[c4d.CAPSANDBEVELS_CAP_TYPE] == c4d.CAPSANDBEVELS_CAP_TYPE_DELAUNAY or \
            #                 node[c4d.CAPSANDBEVELS_CAP_TYPE] == c4d.CAPSANDBEVELS_CAP_TYPE_REGULAR,
            #                 c4d.DESC_PARENT_COLLAPSE)

            # After parameters have been loaded and added successfully, return True and DESCFLAGS_DESC_LOADED with the input flags
            return (True, flags | c4d.DESCFLAGS_DESC_LOADED)

    def IndexUserData(self,obj):
        if obj is None: return False
        grp = ""
        for id, bc in obj.GetUserDataContainer():
            # Material Linkboxes
            if bc[c4d.DESC_CUSTOMGUI] == c4d.CUSTOMGUI_LINKBOX:
                name = bc[c4d.DESC_NAME].split('.')
                field = name[0]
                chksum = name[-1]
                if not chksum in self.userdataIdx:
                    self.userdataIdx[chksum] = {}
                if not field in self.userdataIdx[chksum]:
                    self.userdataIdx[chksum][field] = id
            # Groups
            elif bc[c4d.DESC_CUSTOMGUI] is None:
                if self.matGroup and id != self.matGroup:
                    grp = bc[c4d.DESC_SHORT_NAME]
                    if bc[c4d.DESC_PARENTGROUP] != self.matGroup:
                        grp = grp.split('.')[-1]
                    self.userdataIdx[grp] = {"Grp":id}
                    self.userdataIdx[grp]["Desc"] = bc
                if bc[c4d.DESC_NAME] == "Layers":
                    self.matGroup = id
                else:
                    name = bc[c4d.DESC_SHORT_NAME]
                    if name != "Paths":
                        chksum = name.split('.')[-1]
                        self.userdataIdx[chksum] = {"Grp":id}
                        self.userdataIdx[grp]["Desc"] = bc
            else:
                if grp in self.userdataIdx:
                    if bc[c4d.DESC_SHORT_NAME] == "Offset":
                        self.userdataIdx[grp]["Off"] = id
                    elif bc[c4d.DESC_SHORT_NAME] == "Depth":
                        self.userdataIdx[grp]["Dep"] = id
                    elif bc[c4d.DESC_SHORT_NAME] == "Path Spread":
                        self.userdataIdx[grp]["Pat"] = id
                    elif bc[c4d.DESC_SHORT_NAME] == "Effectors":
                        self.userdataIdx[grp]["Eff"] = id
    
    def CreateUserDataGroup(self, obj, name, parentGroup=None, columns=None, shortname=None):
        if obj is None: return False
        if shortname is None: shortname = name
        bc = c4d.GetCustomDatatypeDefault(c4d.DTYPE_GROUP)
        bc[c4d.DESC_NAME] = name
        bc[c4d.DESC_SHORT_NAME] = shortname
        bc[c4d.DESC_TITLEBAR] = 1
        if parentGroup is not None:
            bc[c4d.DESC_PARENTGROUP] = parentGroup
        if columns is not None:
            bc[c4d.DESC_LAYOUTGROUP] = True
            bc[c4d.DESC_COLUMNS] = columns
            
        return obj.AddUserData(bc)    
    
    def CreateUserDataLink(self, obj, name, link, parentGroup=None, shortname=None):
        if obj is None: return False
        if shortname is None: shortname = name
        bc = c4d.GetCustomDatatypeDefault(c4d.DTYPE_BASELISTLINK)
        bc[c4d.DESC_NAME] = name
        bc[c4d.DESC_SHORT_NAME] = shortname
        bc[c4d.DESC_DEFAULT] = link
        bc[c4d.DESC_ANIMATE] = c4d.DESC_ANIMATE_OFF
        bc[c4d.DESC_SHADERLINKFLAG] = True
        if parentGroup is not None:
            bc[c4d.DESC_PARENTGROUP] = parentGroup

        element = obj.AddUserData(bc)
        obj[element] = link
        return element
    
    def CreateUserDataFloat(self, obj, name, val=0, parentGroup=None, unit=c4d.DESC_UNIT_FLOAT):
        if obj is None: return False
        bc = c4d.GetCustomDatatypeDefault(c4d.DTYPE_REAL)
        bc[c4d.DESC_NAME] = name
        bc[c4d.DESC_SHORT_NAME] = name
        bc[c4d.DESC_DEFAULT] = val
        bc[c4d.DESC_ANIMATE] = c4d.DESC_ANIMATE_ON
        bc[c4d.DESC_UNIT] = unit
        bc[c4d.DESC_CUSTOMGUI] = c4d.CUSTOMGUI_REALSLIDER
        bc[c4d.DESC_MINSLIDER] = -500
        bc[c4d.DESC_MAXSLIDER] = 500
        bc[c4d.DESC_STEP] = 1        
        if parentGroup is not None:
            bc[c4d.DESC_PARENTGROUP] = parentGroup

        element = obj.AddUserData(bc)
        obj[element] = val
        return element
    
    def CreateUserDataInEx(self, obj, name, parentGroup=None):
        if obj is None: return False
        bc = c4d.GetCustomDatatypeDefault(c4d.CUSTOMDATATYPE_INEXCLUDE_LIST)
        bc[c4d.DESC_NAME] = name
        bc[c4d.DESC_SHORT_NAME] = name
        bc[c4d.DESC_CUSTOMGUI] = c4d.CUSTOMGUI_INEXCLUDE_LIST 
        bc[c4d.IN_EXCLUDE_FLAG_NUM_FLAGS] = 1
        bc[c4d.IN_EXCLUDE_FLAG_INIT_STATE] = 1
        bc[c4d.IN_EXCLUDE_FLAG_IMAGE_01_ON] = 1018640
        bc[c4d.IN_EXCLUDE_FLAG_IMAGE_01_OFF] = 1018641
        bc[c4d.DESC_HIDE] = not bool(obj[ART_USEMOGRAPH])
        if parentGroup is not None:
            bc[c4d.DESC_PARENTGROUP] = parentGroup

        element = obj.AddUserData(bc)
        return element

    
    def BuildMaterialUserData(self, node, obj, parentGroup=None, createMaterials=False, outer=False):
        node.GetDocument().AddUndo(c4d.UNDOTYPE_CHANGE_NOCHILDREN,node)

        # Loop through userdataIdx and set a value
        # We'll change this value for each checksum we find
        # So we can delete checksums that no longer exist
        if outer:
            for k in list(self.userdataIdx.keys()):
                self.userdataIdx[k]["live"] = False
              
        # Loop through the children of the object hierarchy
        allPathGroup = None
        for child in obj.GetChildren():
            # print child.GetName()," ",child.GetType()
            # Extrudes get a listing and materials
            if child.GetType() == c4d.Oextrude or child.GetType() == c4d.Osweep:
                
                #Get Checksum from child
                checksum = ""
                tag = child.GetTag(TAG_ID)
                if not tag: tag = child.GetUp().GetTag(TAG_ID)
                if tag: checksum = tag[ART_CHECKSUM]
                
                layerName = child.GetUp().GetName()
                
                if checksum in self.userdataIdx and "Grp" in self.userdataIdx[checksum]:
                    pathGroup = self.userdataIdx[checksum]["Grp"]
                else:
                    pathGroup = self.CreateUserDataGroup(node,"Path."+checksum,parentGroup)
                    self.userdataIdx[checksum] = {"Grp":pathGroup}
                    
                # Create a material for the extrude
                mat = None
                if createMaterials:
                    if node[ART_GREEDYFILL] == 0 and (layerName+str(child[c4d.ID_BASEOBJECT_COLOR])) in self.matIdx:
                        mat = self.matIdx[layerName+str(child[c4d.ID_BASEOBJECT_COLOR])]
                    elif node[ART_GREEDYFILL] == 1 and (checksum+str(child[c4d.ID_BASEOBJECT_COLOR])) in self.matIdx:
                        mat = self.matIdx[checksum+str(child[c4d.ID_BASEOBJECT_COLOR])]
                    else:
                        mat = c4d.BaseMaterial(c4d.Mmaterial)
                        mat[c4d.MATERIAL_COLOR_COLOR] = child[c4d.ID_BASEOBJECT_COLOR]
                        mat[c4d.MATERIAL_LUMINANCE_COLOR] = child[c4d.ID_BASEOBJECT_COLOR]
                        child.GetDocument().InsertMaterial(mat)
                        node.GetDocument().AddUndo(c4d.UNDOTYPE_NEWOBJ,mat)
                        mat.SetName(layerName)
                        if node[ART_GREEDYFILL] == 1:
                            self.matIdx[checksum+str(child[c4d.ID_BASEOBJECT_COLOR])] = mat
                        self.matIdx[layerName+str(child[c4d.ID_BASEOBJECT_COLOR])] = mat
                    # Create userdata
                    #if not allPathGroup:
                    #    allPathGroup = self.CreateUserDataGroup(node,"Paths",parentGroup)
                    #pathGroup = self.CreateUserDataGroup(node,"Path."+checksum,allPathGroup)
                    
                    if child.GetType() == c4d.Oextrude:
                        if "Ext" not in self.userdataIdx[checksum]: 
                            self.userdataIdx[checksum]["Ext"] = self.CreateUserDataLink(node,"Ext."+checksum,mat,pathGroup,shortname="Ext")
                            self.userdataIdx[checksum]["Cap"] = self.CreateUserDataLink(node,"Cap."+checksum,None,pathGroup,shortname="Cap")
                            self.userdataIdx[checksum]["Rnd"] = self.CreateUserDataLink(node,"Rnd."+checksum,None,pathGroup,shortname="Rnd")
                    elif child.GetType() == c4d.Osweep:
                        if "Sweep" not in self.userdataIdx[checksum]:
                            self.userdataIdx[checksum]["Sweep"] = self.CreateUserDataLink(node,"Sweep."+checksum,mat,pathGroup,shortname="Sweep")
 
                # Make sure to indicate that this checksum path exists in the reload
                if checksum in self.userdataIdx:
                    self.userdataIdx[checksum]["live"] = True
                    
            # Nulls or Fractures indicate a layer - we use them as a userdata group for organization
            elif child.GetType() == c4d.Onull or child.GetType() == FRACTURE_ID or self.pathControl:
                name = child.GetName()
                displayName = name
                tag = child.GetTag(TAG_ID)
                if tag: 
                    chksum = tag[ART_CHECKSUM]
                    #displayName = name+"."+chksum
                    name = chksum
                    short = displayName + "." + chksum
                    layer=False
                else:
                    short = displayName
                    layer=True
                             
                allPathGroup=None
                # Set and pass layergroup so children of this null/fracture are grouped under the userdata
                if name in self.userdataIdx and "Grp" in self.userdataIdx[name]:
                    layerGroup = self.userdataIdx[name]["Grp"]
                    if "Desc" in self.userdataIdx[name] and self.userdataIdx[name]["Desc"][c4d.DESC_PARENTGROUP] != parentGroup:
                        self.userdataIdx[name]["Desc"][c4d.DESC_PARENTGROUP] = parentGroup
                        node.SetUserDataContainer(layerGroup,self.userdataIdx[name]["Desc"])
                else:
                    showSweep = True
                    showDepth = True
                    showEff = False
                    layerGroup = self.CreateUserDataGroup(node,displayName,parentGroup,shortname=short)
                    self.userdataIdx[name] = {"Grp":layerGroup}
                    self.userdataIdx[name]["Off"] = self.CreateUserDataFloat(node,"Offset",0,layerGroup,unit=c4d.DESC_UNIT_METER)
                    
                    if layer and (child.GetType() == c4d.Onull or child.GetType() == FRACTURE_ID):
                        self.userdataIdx[name]["Pat"] = self.CreateUserDataFloat(node,"Path Spread",0,layerGroup,unit=c4d.DESC_UNIT_METER)
                        showEff = True
                    else:
                        if tag:
                            showSweep = tag[ART_STROKED]
                            showDepth = tag[ART_FILLED]
                    
                    if showDepth:
                        self.userdataIdx[name]["Dep"] = self.CreateUserDataFloat(node,"Depth",0,layerGroup,unit=c4d.DESC_UNIT_METER)
                    
                    if showEff:
                        self.userdataIdx[name]["Eff"] = self.CreateUserDataInEx(node,"Effectors",layerGroup)
                    
                # Make sure to indicate that this layer exists in the reload
                self.userdataIdx[name]["live"] = True
                if child.GetDown(): self.BuildMaterialUserData(node,child,layerGroup,createMaterials,False)
        
        c4d.StatusSetSpin()
        
        if outer:
            for k in self.userdataIdx:
                if self.userdataIdx[k]["live"] and self.userdataIdx[k]["live"] is False:
                    if "Ext" in self.userdataIdx[k]: node.RemoveUserData(self.userdataIdx[k]["Ext"])
                    if "Cap" in self.userdataIdx[k]: node.RemoveUserData(self.userdataIdx[k]["Cap"])
                    if "Rnd" in self.userdataIdx[k]: node.RemoveUserData(self.userdataIdx[k]["Rnd"])
                    if "Off" in self.userdataIdx[k]: node.RemoveUserData(self.userdataIdx[k]["Off"])
                    if "Dep" in self.userdataIdx[k]: node.RemoveUserData(self.userdataIdx[k]["Dep"])
                    if "Pat" in self.userdataIdx[k]: node.RemoveUserData(self.userdataIdx[k]["Pat"])
                    if "Eff" in self.userdataIdx[k]: node.RemoveUserData(self.userdataIdx[k]["Eff"])
                    if "Grp" in self.userdataIdx[k]: node.RemoveUserData(self.userdataIdx[k]["Grp"])
                    self.userdataIdx[k] = {}
        
            c4d.StatusClear()
        
            node.Message(c4d.MSG_CHANGE)

        return True
    
    def CreateLayers(self, node):
        lyrTop = c4d.documents.LayerObject()
        lyrTop.SetName(os.path.basename(self.fn))
        lyrTop.InsertUnder(node.GetDocument().GetLayerObjectRoot())
        node.SetLayerObject(lyrTop)
        
        for child in node.GetChildren():
            if child.GetType() == c4d.Onull or child.GetType() == FRACTURE_ID:
                lyr = c4d.documents.LayerObject()
                lyr.SetName(child.GetName())
                lyr.InsertUnder(lyrTop)
                self.AssignLayer(child,lyr)
        
    def AssignLayer(self, node, layer, next=True, recurse=True):
        while node:
            node.SetLayerObject(layer)
            if recurse and node.GetDown(): self.AssignLayer(node.GetDown(),layer)
            if next:
                node = node.GetNext()
            else:
                node = None
        return True
    
    def LoadArt(self,node,fn):
        self.loaded = False
        self.fn = fn
        if not os.path.isfile(fn):
            self.loaded = True
            return
        self.art = vecfile.ParseVecFile(fn,not node[ART_FORCEPDF],node[ART_GREEDYFILL])
        if self.art and self.art.msg and self.art.msg[0] != None and len(self.art.msg) > 0:
            node[ART_DEBUGMSG] = "\n---------\n".join(self.art.msg)
        else:
            node[ART_DEBUGMSG] = ""
        self.rebuildUserdata = True
        self.IndexMaterials(node)
        self.loaded = True
        c4d.EventAdd()
    
    def UpdateLayersPanel(self,node):
        if not self.matGroup:
            self.matGroup = self.CreateUserDataGroup(node,"Layers",c4d.DescID(0))
        self.BuildMaterialUserData(node,node.GetCache(),self.matGroup,len(self.matIdx) > 0,True)    
        node.Message(c4d.MSG_CHANGE)
        self.rebuildUserdata = False
        c4d.EventAdd()
    
    def IndexMaterials(self,node):
        for mat in node.GetDocument().GetMaterials():
            self.matIdx[mat.GetName()+str(mat[c4d.MATERIAL_COLOR_COLOR])] = mat
    
    def Message(self, node, type, data):
        #for j,k in c4d.__dict__.iteritems():
        #    if type==k and j[:3]=="MSG":
        #        print j, "[[", type, "]]"
        #        if data: print data
        if type == c4d.MSG_DESCRIPTION_POSTSETPARAMETER:
            # data isn't provided for this message, validate or checkupdate
            # so I'm using a class variable to check if the state changed & force reload
            if node[ART_FORCEPDF] != self.forcePdf or node[ART_GREEDYFILL] != self.greedyFill:
                if node[ART_FILE] and self.loaded:
                    self.LoadArt(node,node[ART_FILE])
                    node.SetDirty(c4d.DIRTYFLAGS_DATA)
                self.forcePdf = node[ART_FORCEPDF]
                self.greedyFill = node[ART_GREEDYFILL]
            # have to do ud enabling here because ud isn't trapped in GetDEnabling
            for item in self.userdataIdx:
                if "Eff" in self.userdataIdx[item]:
                    for id,bc in node.GetUserDataContainer():
                        if id==self.userdataIdx[item]["Eff"]:
                            bc[c4d.DESC_HIDE] = not bool(node[ART_USEMOGRAPH])
                            node.SetUserDataContainer(id,bc)
        elif type == c4d.MSG_DESCRIPTION_COMMAND:
            if data and data['id'][0].id == ART_RELOAD:
                if node[ART_FILE]:
                    self.LoadArt(node,node[ART_FILE])
                    node.SetDirty(c4d.DIRTYFLAGS_DATA)
            elif data and data['id'][0].id == ART_OPEN:
                if node[ART_FILE]:
                    if sys.platform.startswith('darwin'):
                        subprocess.call(('open', node[ART_FILE]))
                    elif os.name == 'nt':
                        os.startfile(node[ART_FILE])
            elif data and data['id'][0].id == ART_CREATEMATERIALS:
                node.GetDocument().StartUndo()

                c4d.StatusSetSpin()
                c4d.StatusSetText("Building Materials")

                if not self.matGroup:
                    self.matGroup = self.CreateUserDataGroup(node,"Layers",c4d.DescID(0))

                # Handle Cineware compatible case (people shouldn't be creating materials in this mode, but they will)
                startobj = node.GetCache()
                if node[ART_CINEWARE]: startobj = node.GetDown()

                self.BuildMaterialUserData(node,startobj,self.matGroup,True,True)

                c4d.StatusClear()

                node.GetDocument().EndUndo()
            elif data and data['id'][0].id == ART_INFO:
                webbrowser.open_new_tab('http://www.cineversity.com/vidplaylist/cv-artsmart/artsmart_download')
            #elif data and data['id'][0].id == c4d.CAPSANDBEVELS_STARTBEVEL_LOADPRESET:
            #    print("here")
        elif type == c4d.MSG_MULTI_MARKMATERIALS:
            if data:
                #not sure how to handle this
                pass
            else:
                #mark all materials linked in ArtSmart's userdata so they aren't deleted
                for id, bc in node.GetUserDataContainer():
                    if bc[c4d.DESC_CUSTOMGUI] == c4d.CUSTOMGUI_LINKBOX:
                        if node[id] and isinstance(node[id],c4d.BaseMaterial):
                            node[id].SetBit(c4d.BIT_MATMARK)
        elif type == c4d.MSG_EDITABLE_END:
            #print node
            node.GetCache().GetDown().SetAbsPos(c4d.Vector(9999))
            # transfer keyframed "live" parameters to objects
            for track in node.GetCTracks():
                descID = track.GetDescriptionID()
                #print descID
                #print descID[0]
                if descID[0].id == ART_DEPTH:
                    self.CopyTrack(track, node.GetCache())
            node.SetDirty(c4d.DIRTYFLAGS_CACHE|c4d.DIRTYFLAGS_CHILDREN)

        elif node.GetCache() is not None and not node.IsDirty(c4d.DIRTYFLAGS_DATA):
            if self.rebuildUserdata is True:
                self.UpdateLayersPanel(node)
                self.rebuildUserdata = False
        return True
    
    def CopyTrack(self, srcTrack, op):
        while op:
            if op.GetType() == c4d.Oextrude:
                newtrk = c4d.CTrack(op, c4d.DescID(c4d.DescLevel(c4d.EXTRUDEOBJECT_MOVE, c4d.DTYPE_VECTOR, 0), c4d.DescLevel(c4d.VECTOR_X, c4d.DTYPE_REAL, 0)))
                newcrv = newtrk.GetCurve()
                src = srcTrack.GetCurve()
                for i in range(0,src.GetKeyCount()-1):
                    newcrv.InsertKey(src.GetKey(i))
                op.InsertTrackSorted(newtrk)
                op.SetDirty(c4d.DIRTYFLAGS_CACHE|c4d.DIRTYFLAGS_CHILDREN)
                #print op.GetFirstCTrack()
            if op.GetDown(): self.CopyTrack(srcTrack, op.GetDown())
            op = op.GetNext()
    
    def Draw(self, op, drawpass, bd, bh):
        if drawpass!= c4d.DRAWPASS_HANDLES:
            return c4d.DRAWRESULT_SKIP

        return c4d.DRAWRESULT_OK

    def GetVirtualObjects(self,op,hh):
        dirty = op.CheckCache(hh) or op.IsDirty(c4d.DIRTYFLAGS_DATA)
        if dirty is False: return op.GetCache(hh)
       
        if op[ART_FILE] and (not self.art or not self.fn == op[ART_FILE]):
            self.LoadArt(op,op[ART_FILE])
        
        if not self.art: return op.GetCache(hh)
        
        artsmartcore.checksums = {}
        self.IndexUserData(op)
        
        fname = os.path.basename(self.fn)
        
        nullType = c4d.Onull
        if c4d.modules.CheckMoGraph() and op[ART_USEMOGRAPH] == True:
            nullType = FRACTURE_ID

        null = c4d.BaseObject(nullType)
        null.SetName(fname)
        if nullType is FRACTURE_ID and op[ART_LAYEREFFECTOR_LIST]:
            null[c4d.ID_MG_MOTIONGENERATOR_EFFECTORLIST] = op[ART_LAYEREFFECTOR_LIST]       
                       
        if self.art.layers:
            for item in self.art.layers:
                self.z = artsmartcore.ParseLayer(layer=item,parent=null,parentLayer=None,nullType=nullType,node=op,userdataIdx=self.userdataIdx,zOff=self.z,printOnly=True)
                if self.z is None: self.z = 0
                self.z -= op[ART_LAYEROFFSET]
        else:
            #layers = {}
            #if art.res and "Properties" in art.res:
            #    for mc,mcd in art.res["Properties"].iteritems():
            #        lyr = c4d.documents.LayerObject()
            #        lyr.SetName(mcd[1])
            #        lyr.InsertUnder(lyrTop)
            #        layers[mc] = lyr
            
            for path in self.art.paths:
                spl,depth,offset,spread = artsmartcore.CreatePath(path,op,userdataIdx=self.userdataIdx)
                spl.InsertUnder(null)
                #if path.markers and path.markers[-1] in layers:
                #    spl.SetLayerObject(layers[path.markers[-1]])
        
        if op[ART_RECENTER] or op[ART_RECENTER_Z]:
            centers = []
            for obj in null.GetChildren():
                centers.append(obj.GetAbsPos())
            if centers:
                if op[ART_LEGACYCENTER]:
                    center = sum(centers)/len(centers)
                else:
                    center = artsmartcore.FindCenter(centers)
                for obj in null.GetChildren():
                    newpos = obj.GetAbsPos()
                    if op[ART_RECENTER]:
                        newpos.x = newpos.x-center.x
                        newpos.y = newpos.y-center.y
                    if op[ART_RECENTER_Z]:
                        newpos.z = newpos.z-center.z
                    obj.SetAbsPos(newpos)
        
        self.z = 0
        
        if op[ART_CINEWARE]:
            #insert under
            null.InsertUnder(op)
            return c4d.BaseObject(c4d.Onull)
        return null

if __name__ == "__main__":
       
    path, file = os.path.split(__file__)
    banner = c4d.bitmaps.BaseBitmap()
    banner.InitWith(os.path.join(path, "res", "banner.psd"))
       
    if banner: c4d.gui.RegisterIcon(ICON_ID, banner)
    
    bmp = c4d.bitmaps.BaseBitmap()
    bmp.InitWith(os.path.join(path, "res", "icon-object.psd"))
    c4d.plugins.RegisterObjectPlugin(id=PLUGIN_ID, str="CV-ArtSmart Object",
                                g=CVArtSmart,
                                description="Ocvartsmart", icon=bmp,
                                info=c4d.OBJECT_GENERATOR|c4d.OBJECT_ISSPLINE|c4d.OBJECT_USECACHECOLOR)
    print("CV-ArtSmart v2.95 // www.cineversity.com/vidplaylist/cv-artsmart")
    c4d.plugins.Update()